!=======================================================================
!                 ___________
!             \\//-----------
!              )(   //== //
!             //\\ //   //
!=======================================================================
!      EXTENDED FORTRAN TYPES LIBRARY ADDONS
!=======================================================================
! XFTExtra.f90 - convenience functions
!
! You are free to use the code for both commercial and non-commercial
! use. The code is provided as-is without warranties. You are free to
! distribute and modify this source code.
!=======================================================================
! An assortment of utility routines by Mike Gaitens for JD's Xeffort framework
!=======================================================================
MODULE XFTextra

USE Xeffort       ! release 16 or later
USE DFWIN

IMPLICIT NONE

      PRIVATE

PUBLIC   XNumber           !(sString, value)
PUBLIC   XSqueeze          !(sString)
PUBLIC   XTimerStart       !(xWnd, idTimer, interval)
PUBLIC   XTimerStop        !(xWnd, idTimer)
PUBLIC   XUserWarn         !(sCaption, sMessage)
PUBLIC   XUserInfo         !(sCaption, sMessage)
PUBLIC   XUserStop         !(sCaption, sMessage)
PUBLIC   XUserAsk          !(sCaption, sMessage, [iBtn], [iDef])
PUBLIC   XSetDialogFocus   !(xWnd, ID)
PUBLIC   XOneInstance      !(unique, [doomed])   ! limits app to single instance
PUBLIC   XPickFont         !(xWnd, xDC/hDC, sFontName, iHeight,    &
                           ! [iEscapement], [iWidth], [iWeight],  &
                           !  [bItalic], [bUnderline], [bStrikeOut], [iFontColor])
PUBLIC   XPickCtlFont      !(xWnd, sCtlFont, [iFontColor])
PUBLIC   XPickColor        !(xWnd, iRGB)

PUBLIC   XPlaySound        ! see below.  plays .wav file or wave resource

      INTERFACE XPlaySound       ! to private routines
         MODULE PROCEDURE PXPlaySound_Rsrc      ! (ID, [hModule])
         MODULE PROCEDURE PXPlaySound_File      ! (sFile)
         MODULE PROCEDURE PXPlaySound_Silence   ! ()
      END INTERFACE


      INTERFACE XNumber          ! to private routines
         MODULE PROCEDURE XNumber_I4            ! (sString, integer)
         MODULE PROCEDURE XNumber_R4            ! (sString, real)
         MODULE PROCEDURE XNumber_R8            ! (sString, doubleprecision)
      END INTERFACE


      INTERFACE XPickFont
         MODULE PROCEDURE XPickFontxDC          ! using xDC
         MODULE PROCEDURE XPickFonthDC          ! using xDC % hDC
      END INTERFACE


!=======================================================================
CONTAINS
!=======================================================================
! XSqueeze removes all embedded spaces and control characters from a string.
!           Any spaces in the return string are trailing spaces.
RECURSIVE FUNCTION XSqueeze (sString) RESULT (s)

      CHARACTER(*), INTENT(IN)           :: sString
      CHARACTER(LEN = LEN_TRIM(sString)) :: s
      INTEGER :: J

      s = sString
!            convert any control characters to spaces
      DO J = 1, LEN (s)
         IF (s(J:J) .LT. ' ') s(J:J) = ' '
      END DO

!          scan s for the position of the leftmost blank
  100 J = INDEX (s, ' ')
      IF (J.LE.0) RETURN
!          done if all to the right of that is also blank
      IF (s(J:) .EQ. ' ') RETURN
!          else shift left to cover the blank and recurse
      s(J:) = s(J+1:)
      GO TO 100

END FUNCTION XSqueeze
!=======================================================================
! XNumber_R8 attempts to read a given text string as a REAL(8) value
      RECURSIVE LOGICAL FUNCTION XNumber_R8 (sString, Value)

      CHARACTER(*), INTENT(IN) :: sString
      REAL(8), INTENT(OUT)     :: Value

!           local store
      CHARACTER(LEN = LEN(sString)) :: s
      INTEGER :: J

      s = XSqueeze(XToUpper (sString))

!           return zero if blank
      IF (s .EQ. ' ') THEN
         Value = 0.0
         XNumber_R8 = .TRUE.          ! success
         RETURN
      ENDIF

!           use D as exponent letter rather than E
  100 J = INDEX (s, 'E')
      IF (J .GT. 0) THEN
         s(J:J) = 'D'
         GO TO 100
      ENDIF
!           attempt conversion
      READ (s, *, ERR = 900) Value
      XNumber_R8 = .TRUE.          ! success

      RETURN

!           conversion failed
  900 XNumber_R8 = .FALSE.
      Value = 0.0

      END FUNCTION XNumber_R8
!=======================================================================
! XNumber_R4 attempts to read a given text string as a REAL(4) value
      RECURSIVE LOGICAL FUNCTION XNumber_R4 (sString, Value)

      CHARACTER(*), INTENT(IN) :: sString
      REAL(4), INTENT(OUT)     :: Value

!           local store
      REAL(8) :: Temp

      XNumber_R4 = XNumber_R8 (sString, Temp)
      Value = Temp
!           insure within range of real*4
      IF (XNumber_R4) THEN
         IF (ABS (Temp) .GT. DBLE (HUGE (Value)) ) THEN
            XNumber_R4 = .FALSE.
            Value = 0.0
         ENDIF
      ENDIF

      END FUNCTION XNumber_R4
!=======================================================================
! XNumber_I4 attempts to read a given text string as a INTEGER(4) value
      RECURSIVE LOGICAL FUNCTION XNumber_I4 (sString, Value)

      CHARACTER(*), INTENT(IN) :: sString
      INTEGER(4), INTENT(OUT)  :: Value

!           local store
      REAL(8) :: Temp

      XNumber_I4 = XNumber_R8 (sString, Temp)
      Value = NINT (Temp)
!           insure within range of integer
      IF (XNumber_I4) THEN
         IF (ABS (Temp) .GT. DBLE (HUGE (Value)) ) THEN
            XNumber_I4 = .FALSE.
            Value = 0
         ENDIF
      ENDIF

      END FUNCTION XNumber_I4
!=======================================================================
!=======================================================================
!           plays a .WAV file encoded in a given resource ID
RECURSIVE LOGICAL FUNCTION PXPlaySound_Rsrc(ID, hModule)

INTEGER, INTENT(IN)::               ID
INTEGER, OPTIONAL, INTENT(IN)::     hModule

INTEGER::                           hApp

IF (PRESENT(hModule)) THEN
      hApp = hModule
ELSE
      hApp = GetModuleHandle(0)
END IF

PXPlaySound_Rsrc = PlaySound (ID, hApp, SND_RESOURCE .OR. SND_NOSTOP   &
                               &                     .OR. SND_ASYNC )

END FUNCTION PXPlaySound_Rsrc

!======================================================================
!           plays the given .WAV file
RECURSIVE LOGICAL FUNCTION PXPlaySound_File(sFile)

CHARACTER(*), INTENT(IN)::          sFile

CHARACTER(LEN = LEN(sFile)+1)::     szFile


szFile = TRIM(sFile)//CHAR(0)

PXPlaySound_File = PlaySound (szFile, 0, SND_FILENAME .OR. SND_NOSTOP  &
                                &                     .OR. SND_ASYNC )

END FUNCTION PXPlaySound_File

!======================================================================
!           stops the playing of any current sound
RECURSIVE LOGICAL FUNCTION PXPlaySound_Silence ()


PXPlaySound_Silence = PlaySound (0, 0, SND_PURGE .OR. SND_NODEFAULT)

END FUNCTION PXPlaySound_Silence
!=======================================================================
!=======================================================================
!XTimerStart creates an interval alarm timer associated with window xWnd.
!The timer is identified by a positive user-assigned ID no.
!The alarm times-out every "interval" milliseconds and generates a WM_TIMER
!message.  It continues timing-out at these regular intervals until it is
!stopped by XTimerStop.
!This function returns .true. if successful.
!

RECURSIVE LOGICAL FUNCTION XTimerStart (xWnd, idTimer, interval)

TYPE(X_WINDOW)::           xWnd        !Parent window of the timer
INTEGER, INTENT(IN)::      idTimer     !user-assigned ID no. of the timer
INTEGER, INTENT(IN)::      interval    !time-out interval in milliseconds

      INTEGER ::        iRet

      XTimerStart = .false.
      IF (idTimer .LE. 0 .OR. interval .LE. 0) RETURN

      iRet = SetTimer (xWnd % hWnd, idTimer, interval, 0)

      XTimerStart = iRet .EQ. idTimer
         IF (iRet .eq. 0) iRet = GetLastError ()

END FUNCTION XTimerStart

!=======================================================================
!XTimerStop destroys an interval timer created by XTimerStart.
!The timer is identified by a positive user-assigned ID no.
!This function returns .true. if successful.
!

RECURSIVE LOGICAL FUNCTION XTimerStop (xWnd, idTimer)

TYPE(X_WINDOW)::           xWnd        !Parent window of the timer
INTEGER, INTENT(IN)::      idTimer     !user-assigned ID no. of the timer

      INTEGER ::        iRet

      XTimerStop = .false.
      IF (idTimer .LE. 0) RETURN

      iRet = KillTimer (xWnd % hWnd, idTimer)

      IF (iRet .NE. 0) THEN
         XTimerStop = .true.
      ELSE
         iRet = GetLastError ()
      ENDIF

END FUNCTION XTimerStop
!=======================================================================
!=======================================================================
!       displays a warning message box with an OK button
SUBROUTINE XUserWarn (sCaption, sMessage, xOwner)

CHARACTER(*), INTENT(IN) :: sCaption, sMessage
TYPE(X_WINDOW), OPTIONAL :: xOwner      ! Parent window of the messagebox

INTEGER::                   iRet

       CALL XMessageBeep (MB_ICONEXCLAMATION)

       iRet = XMessageBox (sMessage, sCaption,                          &
      &                    MB_ICONEXCLAMATION .OR. MB_TOPMOST, xOwner )

END SUBROUTINE XUserWarn
!=======================================================================
!       displays a informatation message box with an OK button
SUBROUTINE XUserInfo (sCaption, sMessage, xOwner)

CHARACTER(*), INTENT(IN) :: sCaption, sMessage
TYPE(X_WINDOW), OPTIONAL :: xOwner      ! Parent window of the messagebox

INTEGER::                   iRet

       CALL XMessageBeep (MB_ICONINFORMATION)

       iRet = XMessageBox (sMessage, sCaption,                          &
      &                    MB_ICONINFORMATION .OR. MB_TOPMOST, xOwner )

END SUBROUTINE XUserInfo
!=======================================================================
!       displays an error message box with an OK button
SUBROUTINE XUserStop (sCaption, sMessage, xOwner)

CHARACTER(*), INTENT(IN) :: sCaption, sMessage
TYPE(X_WINDOW), OPTIONAL :: xOwner      ! Parent window of the messagebox

INTEGER::                   iRet

       CALL XMessageBeep (MB_ICONHAND)

       iRet = XMessageBox (sMessage, sCaption,                          &
      &                    MB_ICONHAND .OR. MB_TOPMOST, xOwner )

END SUBROUTINE XUserStop
!=======================================================================
!     displays a multiple choice message box
RECURSIVE INTEGER FUNCTION XUserAsk (sCaption, sMessage, iBtn, iDef, xOwner)

CHARACTER(*), INTENT(IN) :: sCaption, sMessage
INTEGER, OPTIONAL, INTENT(IN) :: iBtn        ! MB_ specifier of button labels
INTEGER, OPTIONAL, INTENT(IN) :: iDef        ! seq. no. of default button
TYPE(X_WINDOW), OPTIONAL      :: xOwner      ! Parent window of the messagebox

INTEGER::                           iRet, jBtn, jDef

IF (PRESENT (iBtn)) THEN
   SELECT CASE (iBtn)
      CASE (MB_OKCANCEL, MB_ABORTRETRYIGNORE, MB_YESNOCANCEL,       &
          &    MB_YESNO, MB_RETRYCANCEL )
        jBtn = iBtn
      CASE DEFAULT
        jBtn = MB_YESNO
   END SELECT
ELSE
   jBtn = MB_YESNO
ENDIF

IF (PRESENT (iDef)) THEN
   SELECT CASE (iDef)
      CASE (1)
         jDef = MB_DEFBUTTON1
      CASE (2)
         jDef = MB_DEFBUTTON2
      CASE (3)
         SELECT CASE (jBtn)
            CASE (MB_ABORTRETRYIGNORE, MB_YESNOCANCEL )
               jDef = MB_DEFBUTTON3

            CASE (MB_OKCANCEL, MB_YESNO, MB_RETRYCANCEL )
               jDef = MB_DEFBUTTON2
         END SELECT

      CASE DEFAULT
         jDef = MB_DEFBUTTON1
   END SELECT
ELSE
   jDef = MB_DEFBUTTON1
ENDIF


       CALL XMessageBeep (MB_ICONQUESTION)

       jDef = (MB_ICONQUESTION .OR. jBtn .OR. jDef .OR. MB_TOPMOST)

       iRet = XMessageBox (sMessage, sCaption, jDef, xOwner)

!         map ID reply to button placement order
      XUserAsk = 1                   ! presume 1st button hit

      SELECT CASE (jBtn)
         CASE (MB_OKCANCEL, MB_RETRYCANCEL )
            IF (iRet .EQ. IDCANCEL)  XUserAsk = 2

         CASE (MB_YESNO)
            IF (iRet .EQ. IDNO)  XUserAsk = 2

         CASE (MB_ABORTRETRYIGNORE)
            IF (iRet .EQ. IDRETRY)  XUserAsk = 2
            IF (iRet .EQ. IDIGNORE) XUserAsk = 3

         CASE (MB_YESNOCANCEL)
            IF (iRet .EQ. IDNO)      XUserAsk = 2
            IF (iRet .EQ. IDCANCEL)  XUserAsk = 3
      END SELECT

END FUNCTION XUserAsk
!=======================================================================
!=======================================================================
!       switches focus in the given dialog to the control with the given ID
LOGICAL FUNCTION XSetDialogFocus (xWnd, ID)

   TYPE(X_WINDOW)::           xWnd        ! dialog x-window structure
   INTEGER, INTENT(IN)::      ID          ! control id of control in that dialog

      INTEGER :: hCtl                     ! control handle

      XSetDialogFocus = .FALSE.

      hCtl = GetDlgItem(xWnd%hWnd, ID)

      IF (hCtl .NE. 0) THEN
         XSetDialogFocus = PostMessage (xWnd%hWnd, WM_NEXTDLGCTL, hCtl, .TRUE.)
      ENDIF

END FUNCTION XSetDialogFocus

!=======================================================================
!=======================================================================
!              routine to limit application to one instance
      LOGICAL FUNCTION XOneInstance (UNIQUE, DOOMED)
         USE Xeffort
         USE DFWIN

      IMPLICIT NONE

      CHARACTER(*), INTENT(IN) :: UNIQUE        ! user-supplied unique program identifier
      LOGICAL, OPTIONAL, INTENT(OUT) :: DOOMED
!        If present, DOOMED is returned
!           .FALSE. for the primary instance or if an error occurs herein.
!           .TRUE. for a secondary instance that will already have a quit message
!                    posted (so the instance will exit shortly as soon a messages
!                     are processed).

!         The function return value is .TRUE. if successful;
!           or .FALSE. if an error occurred.


!           local store
      CHARACTER(LEN (UNIQUE) +  1) :: UNIQUENUL
      CHARACTER(LEN (UNIQUE) + 16) :: UNIQUEMAP

      INTEGER(HANDLE) :: hMutex
      INTEGER(HANDLE) :: hMainWnd
      INTEGER(HANDLE) :: hMMap

      INTEGER(4) :: J
      LOGICAL :: OK

      INTEGER(4), PARAMETER :: MAXSIZEHI = 0                   ! high 32 bits
      INTEGER(4), PARAMETER :: MAXSIZELO = SIZEOF (hMainWnd)   ! low 32 bits

      INTEGER(POINTER_LEN) :: MapAddr
      INTEGER(HANDLE) :: MapData
      POINTER (MapAddr, MapData)


!           presume failure
      XOneInstance = .FALSE.
      IF (PRESENT (DOOMED) ) DOOMED = .FALSE.

!           don't allow blank arg.
      IF (UNIQUE .EQ. ' ') RETURN

!           make a nul-terminated version of the given arg.
      UNIQUENUL = TRIM (UNIQUE) // CHAR (0)

!           make a name for a memeory map to be shared among instances
      UNIQUEMAP = 'MEMORYMAP' // UNIQUENUL
!           backslashes not allowed in map name so replace any with forward shashes
  200 J = INDEX (UNIQUEMAP, '\' )
      IF (J .GT. 0) THEN
         UNIQUEMAP(J:J) = '/'
         GO TO 200
      ENDIF


!           per JD's suggestion
      CALL SetLastError (0)  !Just in case
      hMutex = CreateMutex (NULL, .TRUE., UNIQUENUL)

!           check success of CreateMutex
      IF (hMutex .EQ. 0) RETURN

!           see if this is the 1st or 2nd instance
      IF (GetLastError () .NE. ERROR_ALREADY_EXISTS) THEN

!           ** this is the 1st instance **

!           get the handle of the main application window
         hMainWnd = XW_FRAME % hWnd
         IF (hMainWnd .EQ. 0) RETURN        ! XCreate___App must have failed

!           create a memory map file for inter-process communication
!             (let file be system page file)
         hMMap = CreateFileMapping  (-1, NULL, PAGE_READWRITE, MAXSIZEHI, MAXSIZELO, UNIQUEMAP)
!              check success
         IF (hMMap .EQ. 0) RETURN

!           get address of mapped memory
         MapAddr = MapViewOfFile (hMMap, FILE_MAP_WRITE, 0, 0, MAXSIZELO)
!              check success
         IF (MapAddr .EQ. 0) RETURN

!           write the handle of our main window to the mapped region
         MapData = hMainWnd

!           ensure value is saved
         IF (FlushViewOfFile (MapAddr, 0) .EQ. 0) RETURN

!           setup complete; mapping file marks us as the only authorized instance
         OK = UnMapViewOfFile (MapAddr)

         XOneInstance = OK

         RETURN

      ELSE

!           ** a second instance of the program is trying to start **

!             gain access to memory mapped region
         hMMap = OpenFileMapping  (FILE_MAP_READ, .TRUE., UNIQUEMAP)
!              check success
         IF (hMMap .EQ. 0) RETURN

!           get address of mapped memory
         MapAddr = MapViewOfFile (hMMap, FILE_MAP_READ, 0, 0, MAXSIZELO)
!              check success
         IF (MapAddr .EQ. 0) RETURN

!           get window handle of the master (1st) instance
         hMainWnd = MapData
         IF (hMainWnd .EQ. 0) RETURN

!           tell the master window to restore itself
         OK = ShowWindow (hMainWnd, SW_SHOWNORMAL)
!              and ensure it's not covered by some other window
         OK = SetForegroundWindow (hMainWnd)

!           we're done with the mapped memory
         OK = UnMapViewOfFile (MapAddr)
         OK = CloseHandle (hMMap)

!           tell our copy of this application to exit
         CALL PostQuitMessage (0)

!           we've done what we set out to do
         XOneInstance = .TRUE.
         IF (PRESENT (DOOMED) ) DOOMED = .TRUE.

      ENDIF


      END FUNCTION XOneInstance
!======================================================================
!======================================================================
!XPickFont launches the ChooseFont common dialog to get font specifications
! for XSetFont
!            returns true/false if user OK's/Cancel's the dialog.
!                      ==================
!        This version for X_DC 2nd argument
      LOGICAL FUNCTION XPickFontxDC (xWnd, xDC, sFontName, iHeight,  &
     &                  iEscapement, iWidth, iWeight,                &
     &                  bItalic, bUnderline, bStrikeOut, iFontColor)

      TYPE (X_WINDOW) ::                xWnd        !parent window for dialog
      TYPE (X_DC),  INTENT(IN)::        xDC         !xDC if printer
      CHARACTER(*), INTENT(OUT)::       sFontName   !Font typeface name
      INTEGER, INTENT(OUT)::            iHeight     !+char Height, in pixels
      INTEGER, OPTIONAL, INTENT(OUT)::  iEscapement !Font angle, in 1/10s of degrees
      INTEGER, OPTIONAL, INTENT(OUT)::  iWidth      !Font width
      INTEGER, OPTIONAL, INTENT(OUT)::  iWeight     !Font weight
      LOGICAL, OPTIONAL, INTENT(OUT)::  bItalic     !Italic font
      LOGICAL, OPTIONAL, INTENT(OUT)::  bUnderline  !Underlined font
      LOGICAL, OPTIONAL, INTENT(OUT)::  bStrikeOut  !Struck-out font
      INTEGER, OPTIONAL, INTENT(OUT)::  iFontColor  !Font RGB color value

      XPickFontxDC = XPickFonthDC (xWnd, xDC % hDC, sFontName, iHeight,  &
     &                  iEscapement, iWidth, iWeight,                &
     &                  bItalic, bUnderline, bStrikeOut, iFontColor)

      END FUNCTION XPickFontxDC

!                      ==================
!        This version for DC handle 2nd argument
      LOGICAL FUNCTION XPickFonthDC (xWnd, hDC, sFontName, iHeight,  &
     &                  iEscapement, iWidth, iWeight,                &
     &                  bItalic, bUnderline, bStrikeOut, iFontColor)

      TYPE (X_WINDOW) ::                xWnd        !parent window for dialog
      INTEGER,  INTENT(IN)::            hDC         !xDC % hDC if printer
      CHARACTER(*), INTENT(OUT)::       sFontName   !Font typeface name
      INTEGER, INTENT(OUT)::            iHeight     !+char Height, in pixels
      INTEGER, OPTIONAL, INTENT(OUT)::  iEscapement !Font angle, in 1/10s of degrees
      INTEGER, OPTIONAL, INTENT(OUT)::  iWidth      !Font width
      INTEGER, OPTIONAL, INTENT(OUT)::  iWeight     !Font weight
      LOGICAL, OPTIONAL, INTENT(OUT)::  bItalic     !Italic font
      LOGICAL, OPTIONAL, INTENT(OUT)::  bUnderline  !Underlined font
      LOGICAL, OPTIONAL, INTENT(OUT)::  bStrikeOut  !Struck-out font
      INTEGER, OPTIONAL, INTENT(OUT)::  iFontColor  !Font RGB color value


!           local store
      INTEGER :: hWnd               ! handle of owner
      TYPE (T_LOGFONT) :: LF
      TYPE (T_CHOOSEFONT) :: CF
      INTEGER :: IFLAG


!           get owner handle
      hWnd = xWnd % hWnd


      IFLAG = CF_EFFECTS .OR. CF_FORCEFONTEXIST

!           if given a DC, consider all fonts for it; else screen fonts only
      IF (hDC .EQ. 0) THEN
         IFLAG = IFLAG .OR. CF_SCREENFONTS
      ELSE
         IFLAG = IFLAG .OR. CF_BOTH
      ENDIF

!           init choosefont structure
      CF % lStructSize = SIZEOF (CF)
      CF % hwndOwner   = hWnd
      CF % hDC         = hDC
      CF % lpLogFont   = LOC (LF)   ! where to put dlg output
      CF % iPointSize  = 0          ! returned from dlg
      CF % Flags       = IFLAG
      CF % rgbColors   = 0          ! init color to black
      CF % lCustData   = 0          ! hook parameter not used
      CF % lpfnHook    = 0          ! hook not used
      CF % lpTemplateName = 0       ! use standard template
      CF % hInstance   = 0          ! use standard template, so not reqd.
      CF % lpszStyle   = 0          ! custom style not used
      CF % nFontType   = 0          ! returned, (and in LF too)
      CF % nSizeMin    = 0          ! no limit applied
      CF % nSizeMax    = 0          ! no limit applied

!           launch the dialog -- returns true upon OK; false if CANCELed
      XPickFonthDC = ChooseFont (CF)

!           set output arguments
      IF (XPickFonthDC) THEN
         sFontName = XC2F (LF % lfFaceName)          !Font typeface name
         iHeight = LF % lfHeight                     !Height, in pixels
!           if iHeight is positive, it is a cell height
!           if iHeight is negative, it is a character height
!           convert cell height to character height if reqd.
!           then flip sign to return character height as a positive iHeight
         IF (iHeight .EQ. 0) iHeight = - 14
         IF (iHeight .GT. 0) THEN
            CALL CHARHTCONVERT (LF, hWnd, hDC, iHeight)
         ENDIF
         iHeight = - iHeight        ! make positive

         IF (PRESENT (iEscapement))  iEscapement = LF % lfEscapement !Font angle, in 1/10s of degrees
         IF (PRESENT (iWidth) )      iWidth      = LF % lfWidth      !Font width
         IF (PRESENT (iWeight) )     iWeight     = LF % lfWeight     !Font weight
         IF (PRESENT (bItalic) )     bItalic     = LF % lfItalic     !Italic font
         IF (PRESENT (bUnderline) )  bUnderline  = LF % lfUnderline  !Underlined font
         IF (PRESENT (bStrikeOut) )  bStrikeOut  = LF % lfStrikeOut  !Struck-out font
         IF (PRESENT (iFontColor) )  iFontColor  = CF % rgbColors    !Font RGB color value
      ENDIF

      END FUNCTION XPickFonthDC

!======================================================================
!CharHtConvert is a private routine to get character height from a logical font
!           specifying cell height
      SUBROUTINE CHARHTCONVERT (LF, hWnd, hDCgvn, iHeight)

      TYPE (T_LOGFONT), INTENT(IN) :: LF        ! logical font
      INTEGER, INTENT(IN)  :: hDCgvn            ! handle of device context
      INTEGER, INTENT(OUT) :: iHeight           ! returned negative

!           local store
      INTEGER :: hWnd               ! handle of owner window
      INTEGER :: hDC                ! handle of device context
      INTEGER :: hFont              ! handle of new font
      INTEGER :: hOrig              ! handle of original font
      INTEGER :: mMode              ! previous map mode

      LOGICAL OK
      INTEGER J

      TYPE (T_TEXTMETRIC) :: TM       ! text metrics
!             TYPE T_TEXTMETRIC
!             SEQUENCE
!               integer(LONG) tmHeight
!               integer(LONG) tmAscent
!               integer(LONG) tmDescent
!               integer(LONG) tmInternalLeading
!               integer(LONG) tmExternalLeading
!               integer(LONG) tmAveCharWidth
!               integer(LONG) tmMaxCharWidth
!               integer(LONG) tmWeight
!               integer(LONG) tmOverhang
!               integer(LONG) tmDigitizedAspectX
!               integer(LONG) tmDigitizedAspectY
!               integer(BYTE) tmFirstChar
!               integer(BYTE) tmLastChar
!               integer(BYTE) tmDefaultChar
!               integer(BYTE) tmBreakChar
!               integer(BYTE) tmItalic
!               integer(BYTE) tmUnderlined
!               integer(BYTE) tmStruckOut
!               integer(BYTE) tmPitchAndFamily
!               integer(BYTE) tmCharSet
!             END TYPE
!
!           if we have a real device context we use it,
!            else use the display context of the window
      IF (hDCgvn .EQ. 0) THEN
         hDC = GetDC (hWnd)
      ELSE
         hDC = hDCgvn
      ENDIF

!           select the given font into the DC and get its font properties
      mMode = SetMapMode (hDC, MM_TEXT)
      hFont = CreateFontIndirect (LF)
      hOrig = SelectObject (hDC, hFont)
      OK = GetTextMetrics (hDC, TM)
      IF (OK) THEN
         iHeight = TM % tmInternalLeading - TM % tmHeight
      ELSE
         iHeight = - ABS (LF % lfHeight)        ! guess if can't get tm
      ENDIF

!          restore original settings
      mMode = SetMapMode (hDC, mMode)
      J = SelectObject (hDC, hOrig)
      OK = DeleteObject (hFont)

      IF (hDCgvn .EQ. 0) THEN
!              the following should do nothing since we expect cs_owndc style
         J = ReleaseDC (hWnd, hDC)
      ENDIF


      END SUBROUTINE CHARHTCONVERT
!======================================================================
!XPickCtlFont launches the ChooseFont common dialog to get font specifications
! for a dialog control to be set via XCtlSet (..., CTL_FONT)
!  and perhaps XCtlSet (..., CTL_COLOR)
!            returns true/false if user OK's/Cancel's the dialog.
      LOGICAL FUNCTION XPickCtlFont(xWnd, sCtlFont, iFontColor)

      TYPE (X_WINDOW) ::               xWnd        !parent window for dialog
      CHARACTER*(*), INTENT(OUT)::     sCtlFont    !compact Font specification
      INTEGER, OPTIONAL, INTENT(OUT):: iFontColor  !Font RGB color value


!           local store
      INTEGER :: hWnd               ! handle of owner
      INTEGER :: iHeight            ! - char height
      TYPE (T_LOGFONT) :: LF
      TYPE (T_CHOOSEFONT) :: CF
      INTEGER :: IFLAG
      CHARACTER(32) :: FACENAME


!           get owner handle
      hWnd = xWnd % hWnd

      IFLAG = CF_FORCEFONTEXIST .OR. CF_SCREENFONTS
      IF (PRESENT (iFontColor) ) IFLAG = IFLAG .OR. CF_EFFECTS

!           init choosefont structure
      CF % lStructSize = SIZEOF (CF)
      CF % hwndOwner   = hWnd
      CF % hDC         = 0
      CF % lpLogFont   = LOC (LF)   ! where to put dlg output
      CF % iPointSize  = 0          ! returned from dlg
      CF % Flags       = IFLAG
      CF % rgbColors   = 0          ! init color to black
      CF % lCustData   = 0          ! hook parameter not used
      CF % lpfnHook    = 0          ! hook not used
      CF % lpTemplateName = 0       ! use standard template
      CF % hInstance   = 0          ! use standard template, so not reqd.
      CF % lpszStyle   = 0          ! custom style not used
      CF % nFontType   = 0          ! returned, (and in LF too)
      CF % nSizeMin    = 0          ! no limit applied
      CF % nSizeMax    = 0          ! no limit applied

!           launch the dialog -- returns true upon OK; false if CANCELed
      XPickCtlFont = ChooseFont (CF)

!           set output argument
      IF (XPickCtlFont) THEN
         FACENAME = XC2F (LF % lfFaceName)          !Font typeface name
         sCtlFont = "t'" // TRIM (FACENAME) // "'"

         iHeight = LF % lfHeight                     !Height, in pixels
!           if iHeight is positive, it is a cell height
!           if iHeight is negative, it is a character height
!           convert cell height to character height if reqd.
!           then flip sign to return character height as a positive iHeight
         IF (iHeight .EQ. 0) iHeight = - 14
         IF (iHeight .GT. 0) THEN
            CALL CHARHTCONVERT (LF, hWnd, 0, iHeight)
         ENDIF
         iHeight = - iHeight        ! make positive
         CALL APPEND ("h" // TWODIGIT (iHeight) )

         IF (LF % lfWidth .GT. 0)  CALL APPEND ("w" // TWODIGIT (LF % lfWidth)  )
         IF (LF % lfWeight .GE. 600)  CALL APPEND ("e")    !Bold font
         IF (LF % lfItalic .NE. 0)    CALL APPEND ("i")    !Italic font
         IF (LF % lfUnderline .NE. 0) CALL APPEND ("u")    !Underlined font
         IF (PRESENT (iFontColor) )  iFontColor  = CF % rgbColors    !Font RGB color value
      ENDIF

      RETURN

!           ==========
      CONTAINS
!           ==========
!             appends arg to sCtlFont
         SUBROUTINE APPEND (ARG)
            CHARACTER(*) ARG
            INTEGER L

            L = LEN_TRIM (sCtlFont)
            sCtlFont(L+1:) = ARG

            RETURN
         END SUBROUTINE APPEND

!           translates integer arg to 2-digit text
         FUNCTION TWODIGIT (IARG)  RESULT (STR)
            INTEGER IARG
            CHARACTER(2) STR

            INTEGER J, IERR

            J = ABS (IARG)
            IF (J .GE. 100) J = 99

            WRITE (STR, '(I2.2)', IOSTAT = IERR) J

            RETURN
         END FUNCTION TWODIGIT

      END FUNCTION XPickCtlFont

!===================================================================
! XPickColor invokes common color selection dialog
!            returns true/false if user OK's/Cancel's the dialog.
      LOGICAL FUNCTION XPickColor (xWnd, iRGB)

      TYPE (X_WINDOW) ::        xWnd        ! parent window for dialog
      INTEGER, INTENT(INOUT) :: iRGB        ! color value updated herein
!
!
!          local store
      TYPE (T_CHOOSECOLOR) :: chsclr
!          storage of user's custom color array
      INTEGER, SAVE :: CustomRGB(16) = Z'00FFFFFF'
!
      iRGB = IAND (iRGB, Z'00FFFFFF')       ! insure arg is a legal color
!
!          set T_CHOOSECOLOR structure
      chsclr % lStructSize = SIZEOF (chsclr)
      chsclr % hwndOwner = xWnd % hWnd
      chsclr % hInstance = 0
      chsclr % rgbResult = iRGB                           ! default = original
      chsclr % lpCustColors = LOC (CustomRGB)
      chsclr % Flags = CC_ANYCOLOR .OR. CC_FULLOPEN .OR. CC_RGBINIT
      chsclr % lCustData = 0      ! no hook so no hook data
      chsclr % lpfnHook = 0       ! no hook
      chsclr % lpTemplateName = 0 ! use std dialog

!             invoke standard color selector dialog
      XPickColor = ChooseColor ( chsclr )
!
      IF ( XPickColor ) THEN
!                 user hit ok, return new color
         iRGB = chsclr % rgbResult
      ENDIF
!
      RETURN

      END FUNCTION XPickColor

!======================================================================
!=======================================================================
END MODULE XFTextra
